--------------------------------------------------------------------------------
-- BINARY.E  v0.10  alpha
-- (c) 1999 Gabriel Boehme (gboehme@musicland.com)
--------------------------------------------------------------------------------

-- Eu include files
include file.e
include get.e
include machine.e

-- GB include files
include gbstring.e

-- abend routine
procedure abend(sequence s)
   clear_screen()
   printf(2, "BINARY.E abend: %s\n", {s})
   ? 1/0
end procedure

-- general-purpose types
type boolean(integer x)
   return x = 0 or x = 1
end type

type natural(integer x)
   return x >= 1
end type

type whole(integer x)
   return x >= 0
end type

--------------------------------------------------------------------------------
-- binary layout building blocks & basic data types
--------------------------------------------------------------------------------

sequence bTypes, bBytes, getFunc, putFunc
bTypes  = {}
bBytes  = {}
getFunc = {}
putFunc = {}

type btype(integer x)
   return x >= 1 and x <= length(bTypes)
end type

global constant VarLen = -1, NoRoutine = -1

type btiolen(integer x)
   return x >= 1 or x = VarLen
end type

global function newBTypeIO(btiolen bytes, integer get_rid, integer put_rid)
-- define a new type to convert directly to & from a sequence of bytes
   bTypes  &= 0
   bBytes  &= bytes
   getFunc &= get_rid
   putFunc &= put_rid
   return length(bTypes)
end function

function get_sbyte(integer x)
   if x >= #80 then
      return x - #100
   end if
   return x
end function

function get_uint16(sequence s)
   return s[1] + s[2] * #100
end function

function get_sint16(sequence s)
integer x
   x = s[1] + s[2] * #100
   if x >= #8000 then
      return x - #10000
   end if
   return x
end function

function put_int16(integer x)
   return {remainder(x,#100), floor(x/#100)}
end function

function get_sint32(sequence s)
atom a
   a = bytes_to_int(s)
   if a >= #80000000 then
      return a - #100000000
   end if
   return a
end function

-- basic binary data types
global constant
 uByte = newBTypeIO(1, NoRoutine, NoRoutine)
,sByte = newBTypeIO(1, routine_id("get_sbyte"), NoRoutine)

,uInt16 = newBTypeIO(2, routine_id("get_uint16"), routine_id("put_int16"))
,sInt16 = newBTypeIO(2, routine_id("get_sint16"), routine_id("put_int16"))
,uWord  = uInt16
,sWord  = sInt16

,uInt32 = newBTypeIO(4, routine_id("bytes_to_int"), routine_id("int_to_bytes"))
,sInt32 = newBTypeIO(4, routine_id("get_sint32"), routine_id("int_to_bytes"))
,uDWord = uInt32
,sDWord = sInt32

,Float32 = newBTypeIO(4, routine_id("float32_to_atom"), routine_id("atom_to_float32"))
,Float64 = newBTypeIO(8, routine_id("float64_to_atom"), routine_id("atom_to_float64"))

function get_stringspaces(sequence s)
-- based on the rtrim() routine by Jiri Babor
integer right
   right = length(s)
   while right >= 1 do
      if s[right] = ' ' then
	 right -= 1
      else
	 exit
      end if
   end while
   return s[1..right]
end function

function put_stringspaces(sequence s, integer field_len)
integer len
   len = length(s)
   if len >= field_len then
      -- truncate to fit
      return s[1..field_len]
   end if
   return s & repeat(' ', field_len - len)
end function

function get_stringnull(sequence s)
integer f
   f = find(0, s)
   if f then
      return s[1..f-1]
   end if
   return s
end function

function put_stringnull(sequence s, integer field_len)
integer len
   len = length(s)
   if len >= field_len then
      -- truncate to fit
      return s[1..field_len]
   end if
   return s & repeat(0, field_len - len)
end function

constant
 AreaID = newBTypeIO(VarLen, NoRoutine, NoRoutine)
,StringSpacesID = newBTypeIO(VarLen, routine_id("get_stringspaces"), routine_id("put_stringspaces"))
,StringNullID = newBTypeIO(VarLen, routine_id("get_stringnull"), routine_id("put_stringnull"))

global function VarLenField(btype id, btiolen x)
   if x > 0 then
      return {id, x}
   end if
   return id
end function

global function Area(integer x)
   return VarLenField(AreaID, x)
end function

global function StringSpaces(integer x)
   return VarLenField(StringSpacesID, x)
end function

global function StringNull(natural x)
   return VarLenField(StringNullID, x)
end function

type btype_or_bstring(object x)
   if integer(x) and x >= 1 and x <= length(bTypes) then
      return 1
   elsif sequence(x) and length(x) = 2 and btype(x[1]) and natural(x[2]) then
      return 1
   end if
   return 0
end type

integer id

global function startBinaryType()
-- define a new type to be made of other types
   bTypes  = append(bTypes, {})
   bBytes  &= 0
   getFunc &= NoRoutine
   putFunc &= NoRoutine
   id = length(bTypes)
   return id
end function

global function addField(natural n, btype_or_bstring field_def)
sequence s
integer field
   if id = 0 then
      abend("addField() must use startBinaryType() first")
   end if
   
   if sequence(field_def) then
      field = field_def[1]
   else
      field = field_def
   end if
   
   if field >= id then
      -- no recursive layouts allowed!
      abend("addField() possible attempt to define recursive type")
   elsif bBytes[id] != 0 then
      -- can't add any more fields to this type
      abend("addField() attempt to add field to a closed type")
   end if
   
   s = append(bTypes[id], n & field_def)
   bTypes[id] = s
   return length(s)
end function

global procedure endBinaryType()
sequence bt, s
integer field, len
   
   if id = 0 then
      abend("endBinaryType() must use startBinaryType() first")
   end if
   
   bt = bTypes[id]
   for i = 1 to length(bt) do
      s = bt[i]
      
      field = s[2]
      len = length(s)
      if len = 3 then
	 if bBytes[field] != VarLen then
	    -- shouldn't have 
	    abend("endBinaryType() attempt to define length for fixed-length field type")
	 end if
	 bBytes[id] += s[1] * s[3]
      elsif len = 2 then
	 if bBytes[field] = VarLen then
	    -- variable-length type
	    bBytes[id] = VarLen
	    exit
	 end if
	 bBytes[id] += s[1] * bBytes[field]
      else
	 abend("endBinaryType() invalid field definition!")
      end if
   end for
   
   id = 0
end procedure

global procedure setBinaryTypeIO(btype id, integer get_rid, integer put_rid)
   getFunc[id] = get_rid
   putFunc[id] = put_rid
end procedure


constant
 ByteStringID = startBinaryType()
,ByteStringByte = addField(1, uByte)
,ByteStringArea = addField(1, StringSpaces(VarLen))
		endBinaryType()

constant
 WordStringID = startBinaryType()
,WordStringWord = addField(1, uWord)
,WordStringArea = addField(1, StringSpaces(VarLen))
		endBinaryType()

if WordStringWord or WordStringArea then
end if

function get_lenstring(sequence s)
integer len
   len = s[ByteStringByte]
   s = s[ByteStringArea]
   if len >= length(s) then
      return s
   end if
   return s[1..len]
end function

function put_lenstring(sequence s, integer str_len)
integer len
   len = length(s)
   if len > str_len then
      len = str_len
      s = s[1..str_len]
   end if
   return {len, s}
end function

setBinaryTypeIO(ByteStringID, routine_id("get_lenstring"), routine_id("put_lenstring"))
setBinaryTypeIO(WordStringID, routine_id("get_lenstring"), routine_id("put_lenstring"))

global function ByteString(integer len)
   return VarLenField(ByteStringID, len)
end function

global function WordString(integer len)
   return VarLenField(WordStringID, len)
end function

global function newRecord(btype id)
sequence s
   s = bTypes[id]
   for i = 1 to length(s) do
      if length(s[i]) = 3 then
	 s[i] = ""
      else
	 s[i] = 0
      end if
   end for
   return s
end function

--------------------------------------------------------------------------------
-- binary file get & put routines
--------------------------------------------------------------------------------

integer fn, return_code

function getBytes(natural bytes)
integer c
sequence s
   if bytes = 1 then
      c = getc(fn)
      if c = -1 then
	 return_code = GET_EOF
      end if
      return c
   end if
   
   s = get_bytes(fn, bytes)
   if length(s) != bytes then
      return_code = GET_EOF
   end if
   return s
end function

function getMe(natural n, btype id, integer bytes)
-- yes, this is REALLY icky code -- but it works!
integer b
sequence s, bt
   
   s = repeat(0, n)
   for i = 1 to n do
      if sequence(bTypes[id]) then
	 bt = bTypes[id]
	 for j = 1 to length(bt) do
	    if length(bt[j]) = 3 then
	       b = bt[j][3]
	       bt[j] = getMe(bt[j][1], bt[j][2], b)
	    else
	       b = bBytes[bt[j][2]]
	       bt[j] = getMe(bt[j][1], bt[j][2], bytes)
	    end if
	    bytes -= b
	 end for
	 s[i] = bt
      else
	 if bBytes[id] > 0 then
	    bytes = bBytes[id]
	 end if
	 s[i] = getBytes(bytes)
      end if
      
      if return_code != GET_SUCCESS then
	 return s
      elsif getFunc[id] != -1 then
	 s[i] = call_func(getFunc[id], {s[i]})
      end if
   end for
   
   if n = 1 then
      return s[1]
   end if
   return s
end function

global function get_binary(integer file, btype_or_bstring x)
object o
   fn = file
   return_code = GET_SUCCESS
   if sequence(x) then
      o = getMe(1, x[1], x[2])
   else
      o = getMe(1, x, 0)
   end if
   return {return_code, o}
end function


procedure putBytes(object bytes)
   puts(fn, bytes)
--    return not equal(bytes, and_bits(#FF, floor(bytes)))
end procedure

procedure putMe(natural n, btype id, integer bytes, object data)
-- another contestant for the "really icky code" award...
integer b
sequence s, bt
   if n = 1 then
      s = {data}
   else
      s = data
   end if
   
   if n != length(s) then
      abend(sprintf("putMe() data list does not match n length (%d != %d)", {length(s), n}))
   end if
   
   for i = 1 to n do
      if bBytes[id] > 0 then
	 bytes = bBytes[id]
	 if putFunc[id] != -1 then
	    s[i] = call_func(putFunc[id], {s[i]})
	 end if
      else
	 if putFunc[id] != -1 then
	    s[i] = call_func(putFunc[id], {s[i], bytes})
	 end if
      end if
      
      if sequence(bTypes[id]) then
	 bt = bTypes[id]
	 if atom(s[i]) or length(s[i]) != length(bt) then
	    abend("putMe() data does not match binary type")
	 end if
	 for j = 1 to length(bt) do
	    if length(bt[j]) = 3 then
	       b = bt[j][3]
	       putMe(bt[j][1], bt[j][2], b,     s[i][j])
	    else
	       b = bBytes[bt[j][2]]
	       putMe(bt[j][1], bt[j][2], bytes, s[i][j])
	    end if
	    bytes -= b
	 end for
      else
	 if (sequence(s[i]) and length(s[i]) != bytes)
	 or (atom(s[i])     and bytes != 1) then
	    abend("putMe() incorrect number of bytes for this type")
	 end if
	 putBytes(s[i])
      end if
   end for
end procedure

global procedure put_binary(integer file, btype_or_bstring x, object data)
   fn = file
   return_code = GET_SUCCESS
   if sequence(x) then
      putMe(1, x[1], x[2], data)
   else
      putMe(1, x, 0, data)
   end if
end procedure


--------------------------------------------------------------------------------
-- binary random-access file routines
--------------------------------------------------------------------------------

sequence FileList
FileList = {}

type random_file(integer fn)
   fn += 1
   if fn >= 1 and fn <= length(FileList) then
      return sequence(FileList[fn])
   end if
   return 0
end type

global function open_random(sequence file, sequence mode, btype id)
integer fn, max_records
sequence s
   
   if length(mode) = 1 then
      -- all random files are binary
      mode &= 'b'
   end if
   
   fn = open(file, mode)
   if fn != -1 then
      if bBytes[id] = VarLen then
	 -- can't open, must use fixed-length type
	 close(fn)
	 abend("open_random() cannot use variable-length binary type")
      end if
      
      s = dir(file)
      s = s[1]
      if remainder(s[D_SIZE], bBytes[id]) then
	 -- can't open, file/record length is incompatible with binary type
	 close(fn)
	 --return -1
	 abend("open_random() binary type length incompatible with file size")
      end if
      max_records = s[D_SIZE] / bBytes[id]
      
      fn += 1
      if fn > length(FileList) then
	 FileList &= repeat(0, fn-length(FileList))
      end if
      FileList[fn] = {id, max_records}
      fn -= 1
   end if
   
   return fn
end function

constant FileType = 1, FileMaxRecs = 2

global function get_random(random_file file, natural rn)
sequence info
object o
   fn = file
   info = FileList[fn+1]
   if rn > info[FileMaxRecs] then
      return {GET_EOF, 0}
   end if
   
   if seek(fn, (rn-1) * bBytes[info[FileType]]) then
      return {GET_FAIL, 0}
   end if
   
   return_code = GET_SUCCESS
   o = getMe(1, info[FileType], 0)
   return {return_code, o}
end function

global procedure put_random(random_file file, natural rn, sequence data)
sequence info
   fn = file
   info = FileList[fn+1]
   if seek(fn, (rn-1) * bBytes[info[FileType]]) then
      abend("put_random() seek failed")
   end if
   
   putMe(1, info[FileType], 0, data)
end procedure

global procedure close_random(random_file fn)
   FileList[fn+1] = 0
   close(fn)
end procedure

--------------------------------------------------------------------------------
-- Revision history:
-- -----------------
-- 
-- v0.10  first release
-- 
-- v0.01  MAJOR CHANGES!!!
--  thru  rewrote all the binary data-handling stuff to use routine_id
-- v0.09  rewrote get & put routines to take these changes into account
--        modified random-access routines to account for these changes
--        fixed problems arising from all these changes
--        belatedly started keeping track of version numbers
--        tested & debugged w/ various binary files
--        FINALLY GOT IT TO WORK!
--       
-- v0.00  created, achieved very basic (& inconsistent) functionality
-- 
--------------------------------------------------------------------------------
-- Thanks to:
-- ----------
-- 
-- 
-- Did I forget you? Send me an e-mail and let me know!
-- 
--------------------------------------------------------------------------------

