-- gbseq.e  v1.0
-- (c) 1999 Gabriel Boehme (gboehme@musicland.com)
--          except where noted

global function insert(sequence s, integer pos, object x)
-- insert object x at position pos of sequence s
-- (to insert at the end of a sequence, pos should be length(s)+1
integer len
   
   -- add a new position at the end of the sequence
   s = append(s, 0)
   len = length(s)
   
   -- verify the insert-to position
   if pos > len or pos < 1 then
      printf(2, "Attempt to insert at position %d of %d-length sequence\n",
		{pos, len-1})
      ? 1/0
   end if
   
   -- scoot the data over and plunk our new value in at pos
   s[pos+1..len] = s[pos..len-1]
   s[pos] = x

   return s
end function

global function delete(sequence s, integer pos)
-- delete position pos from sequence s
integer len
   len = length(s)
   
   -- verify the delete position
   if pos > len or pos < 1 then
      printf(2, "Attempt to delete at position %d of %d-length sequence\n", 
		{pos, len})
      ? 1/0
   end if
   
   -- scoot the data back to pos and delete from the end of the sequence
   -- this effectively deletes the data at position pos
   s[pos..len-1] = s[pos+1..len]
   return s[1..len-1]
end function

global function find_all(object value, sequence s)
-- return a list of all indexes where value is found in s
-- [this routine originally written by Hawke']
sequence list
   list = {}
   
   for i = 1 to length(s) do
      if equal(value, s[i]) then
	 list = append(list, i)
      end if
   end for
   
   return list
end function

global function match_all(sequence value, sequence s)
-- return a list of all indexes where value matches a substring of s
integer m
sequence list
   list = {}
   
   m = match(value, s)
   while m do
      list = append(list, m)
      s[m] = not s[m]
      m = match(value, s)
   end while
   
   return list
end function

global function shuffle(sequence s)
-- randomly shuffle the elements of a sequence
--    based Jiri Babor's Euphorian interpretation of the "uniformly random"
--    1963 shuffle algorithm written by L.E. Moses and R.V. Oakford
integer r
object temp
    for i = length(s) to 2 by -1 do
	r = rand(i)
	temp = s[r]
	s[r] = s[i]
	s[i] = temp
    end for
    return s
end function


