-- gbstring.e  v1.0
-- (c) 1999 Gabriel Boehme (gboehme@musicland.com)
--          except where noted

include gbseq.e

global function char_in_range(integer c, integer lo, integer hi)
   if c < lo then
      return 0
   elsif c > hi then
      return 0
   end if

   return 1
end function

global function string_in_range(sequence s, integer lo, integer hi)
object x
   for i = 1 to length(s) do
      x = s[i]
      if integer(x) then
	 if x < lo then
	    return 0
	 elsif x > hi then
	    return 0
	 end if
      else
	 return 0
      end if
   end for
   
   return 1
end function

type char(integer c)
   return char_in_range(c, 0, #FF)
end type

type string(sequence s)
   return string_in_range(s, 0, #FF)
end type

type string_list(sequence s)
   for i = 1 to length(s) do
      if not string(s[i]) then
	 return 0
      end if
   end for
   return 1
end type

constant WHITESPACE = " \n\t\r"

global function rtrim(string s)
-- trim whitespace from the right side of string s
--    modified Jiri Babor code
integer right
   right = length(s)
   while right >= 1 do
      if find(s[right], WHITESPACE) then
	 right -= 1
      else
	 exit
      end if
   end while
   return s[1..right]
end function

global function trim(string s)
-- trim whitespace from both ends of string s
--    modified Jiri Babor code
integer left, right
   
   -- right trim
   right = length(s)
   while right >= 1 do
      if find(s[right], WHITESPACE) then
	 right -= 1
      else
	 exit
      end if
   end while
   
   -- left trim
   left = 1
   while left <= right do
      if find(s[left], WHITESPACE) then
	 left += 1
      else
	 exit
      end if
   end while
   
   return s[left..right]
end function

global function from_to(integer begin_value, integer end_value)
integer increment
sequence s
   
   if begin_value > end_value then
      increment = -1
   else
      increment = 1
   end if
   
   s = {}
   for value = begin_value to end_value by increment do
      s = append(s, value)
   end for
   
   return s
end function

global function parse(string s, char c)
-- parse string s based on delimiter c into a sequence of values
sequence list
integer prev, curr
   
   if length(s) then
      -- we append x to s so we have at least one delimiter at the end
      list = find_all(c, s & c) -- list of delimiter positions
      prev = 0
      for i = 1 to length(list) do
	 curr = list[i]
	 list[i] = s[prev+1..curr-1]
	 prev = curr
      end for
   
      return list -- now, a list of the delimited values
   end if
   
   return {}
end function

global function deparse(string_list list, char c)
-- convert parsed list back into a string with c delimiters
sequence s
integer len
   len = length(list)
   
   if len then
      s = list[1]
      for i = 2 to len do
	 s = s & c & list[i]
      end for
      
      return s
   end if
   
   return ""
end function


